import logging

from django.conf import settings

__all__ = [
    "auto_approve",
    "dashboard_public",
    "domain_name",
    "get_config",
    "max_retries",
    "retry_interval",
    "timeout",
    "url_scheme",
    "use_celery",
]


NAMESPACE = "WEBMENTIONS"
SETTING_USE_CELERY = f"{NAMESPACE}_USE_CELERY"
SETTING_AUTO_APPROVE = f"{NAMESPACE}_AUTO_APPROVE"
SETTING_URL_SCHEME = f"{NAMESPACE}_URL_SCHEME"
SETTING_TIMEOUT = f"{NAMESPACE}_TIMEOUT"
SETTING_MAX_RETRIES = f"{NAMESPACE}_MAX_RETRIES"
SETTING_RETRY_INTERVAL = f"{NAMESPACE}_RETRY_INTERVAL"
SETTING_DASHBOARD_PUBLIC = f"{NAMESPACE}_DASHBOARD_PUBLIC"

"""settings.DOMAIN_NAME is sometimes used by other libraries for the same purpose,
no need to lock it to our namespace."""
SETTING_DOMAIN_NAME = "DOMAIN_NAME"

DEFAULTS = {
    SETTING_DOMAIN_NAME: None,
    SETTING_AUTO_APPROVE: False,
    SETTING_URL_SCHEME: "https",
    SETTING_USE_CELERY: True,
    SETTING_TIMEOUT: 10,
    SETTING_MAX_RETRIES: 5,
    SETTING_RETRY_INTERVAL: 60 * 10,
    SETTING_DASHBOARD_PUBLIC: False,
}


log = logging.getLogger(__name__)


def _get_attr(key: str):
    return getattr(settings, key, DEFAULTS[key])


def get_config() -> dict:
    return {key: _get_attr(key) for key in DEFAULTS.keys()}


def domain_name() -> str:
    """Return settings.DOMAIN_NAME."""
    return _get_attr(SETTING_DOMAIN_NAME)


def use_celery() -> bool:

    """Return settings.WEBMENTIONS_USE_CELERY, or True if not set.

    This setting enables/disables the use of `celery` for running tasks.
    If disabled, user must run these tasks using `manage.py pending_mentions` management command."""
    return _get_attr(SETTING_USE_CELERY)


def auto_approve() -> bool:
    """Return settings.WEBMENTIONS_AUTO_APPROVE, or False if not set.

    If True, any received Webmentions will immediately become 'public': they will be included in `/get` API responses.
    If False, received Webmentions must be approved by a user with `approve_webmention` permissions.
    """
    return _get_attr(SETTING_AUTO_APPROVE)


def timeout() -> float:
    """Return settings.WEBMENTIONS_TIMEOUT.

    Timeout (in seconds) used for network requests when sending or verifying webmentions."""
    return _get_attr(SETTING_TIMEOUT)


def max_retries() -> int:
    """Return settings.WEBMENTIONS_MAX_RETRIES.

    We may retry processing of webmentions if the remote server is inaccessible.
    This specifies the maximum number of retries before we give up."""
    return _get_attr(SETTING_MAX_RETRIES)


def retry_interval() -> int:
    """Return settings.WEBMENTIONS_RETRY_INTERVAL.

    We may retry processing of webmentions if the remote server is inaccessible.
    This specifies the delay (in seconds) between attempts.

    If using `celery`, this should be precise. Otherwise, this will be treated
    as a minimum interval (and will also depend on `cron` or whatever you are
    using to schedule mention processing).

    Warning: If using RabbitMQ, it may need to be reconfigured if you want to
    specify an interval of more than 15 minutes. See
    https://docs.celeryq.dev/en/stable/userguide/calling.html#eta-and-countdown
    for more information.
    """
    return _get_attr(SETTING_RETRY_INTERVAL)


def url_scheme() -> str:
    """Return settings.WEBMENTIONS_URL_SCHEME.

    This defaults to `https` which is hopefully what your server is using.
    It's handy to be able to choose when debugging stuff though."""
    scheme = _get_attr(SETTING_URL_SCHEME)
    if not settings.DEBUG and scheme != "https":
        log.warning(
            f"settings.{SETTING_URL_SCHEME} should not be `http` when in production!"
        )

    return scheme


def dashboard_public() -> bool:
    """Return settings.WEBMENTIONS_DASHBOARD_PUBLIC.

    Intended for"""
    is_dashboard_public = _get_attr(SETTING_DASHBOARD_PUBLIC)
    if not settings.DEBUG and is_dashboard_public:
        log.warning(
            f"settings.{SETTING_DASHBOARD_PUBLIC} should not be `True` when in production!"
        )

    return is_dashboard_public



from enum import Enum
from typing import List, Tuple

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import timezone

__all__ = [
    "IncomingMentionType",
    "QuotableMixin",
]


class IncomingMentionType(Enum):
    """Properties that describe the context of the incoming webmention.

    See: https://microformats.org/wiki/h-entry"""

    Bookmark = "u-bookmark-of"
    Like = "u-like-of"
    Listen = "u-listen-of"
    Reply = "u-in-reply-to"
    Repost = "u-repost-of"
    Translation = "u-translation-of"
    Watch = "u-watch-of"

    @classmethod
    def choices(cls) -> List[Tuple[str, str]]:
        return [(x.lower(), x) for x in cls.__members__.keys()]


class QuotableMixin(models.Model):
    """A third-party thing that talks about a first-party thing."""

    class Meta:
        abstract = True

    target_url = models.URLField(help_text="Our URL that is mentioned")
    source_url = models.URLField(help_text="The URL that mentions our content")

    quote = models.CharField(
        max_length=300,
        blank=True,
        null=True,
        help_text="A short excerpt from the quoted piece",
    )

    post_type = models.CharField(
        max_length=64,
        null=True,
        blank=True,
        help_text="Type (e.g. like, reply, etc.) of mention, if specified",
        choices=IncomingMentionType.choices(),
    )

    published = models.DateTimeField(default=timezone.now)

    hcard = models.ForeignKey(
        "HCard",
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )

    # The mention target may be an instance of any class that
    # uses the MentionableMixin so we need to use a GenericForeignKey
    content_type = models.ForeignKey(ContentType, null=True, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField(blank=True, null=True)
    target_object = GenericForeignKey("content_type", "object_id")

    def __str__(self):
        return (
            f"{_trim_to_length(self.source_url)} -> {_trim_to_length(self.target_url)}"
        )


def _trim_to_length(text: str, max_length=32) -> str:
    if len(text) >= max_length:
        return f"{text[:max_length - 1]}…"
    else:
        return text



import logging
import re
from typing import Optional, Tuple
from urllib.parse import urlsplit

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator
from requests import RequestException, Response

from mentions import options
from mentions.exceptions import TargetNotAccessible
from mentions.models import OutgoingWebmentionStatus
from mentions.resolution import get_or_create_outgoing_webmention
from mentions.util import html_parser, http_get, http_post

__all__ = [
    "try_send_webmention",
]


STATUS_MESSAGE_TARGET_UNREACHABLE = "The target URL could not be retrieved: {error}."
STATUS_MESSAGE_TARGET_ERROR_CODE = (
    "The target URL returned an HTTP error code: {status_code}."
)
STATUS_MESSAGE_TARGET_ENDPOINT_ERROR = (
    "The target endpoint URL returned an HTTP error code"
)
STATUS_MESSAGE_OK = "The target server accepted the webmention."

log = logging.getLogger(__name__)


def try_send_webmention(
    source_urlpath: str,
    target_url: str,
    outgoing_status: Optional[OutgoingWebmentionStatus] = None,
) -> Optional[bool]:
    """Try to send a webmention for target_url.

    Returns:
        True if a webmention was submitted successfully.
        False if a webmention endpoint was resolved but submission failed.
        None if a webmention endpoint could not be resolved (i.e. the website does not appear to support webmentions).
    """
    if outgoing_status is None:
        outgoing_status = get_or_create_outgoing_webmention(source_urlpath, target_url)

    try:
        response = _get_target(outgoing_status, target_url)
    except TargetNotAccessible:
        return

    endpoint = _get_absolute_endpoint_from_response(response)
    if endpoint:
        log.debug(f"Found webmention endpoint: '{endpoint}'")
        return _try_send_webmention(
            outgoing_status,
            source_urlpath=source_urlpath,
            endpoint=endpoint,
            target_url=target_url,
        )

    else:
        log.info(f"No webmention endpoint found for url '{target_url}'")


def _get_target(
    status: OutgoingWebmentionStatus,
    target_url: str,
) -> Optional[Response]:
    """Confirm the target is accessible."""
    log.debug(f"Checking url='{target_url}' for webmention support...")
    try:
        response = http_get(target_url)
        if response.status_code < 300:
            return response

        error_message = STATUS_MESSAGE_TARGET_ERROR_CODE.format(
            status_code=response.status_code
        )

    except RequestException as e:
        error_message = STATUS_MESSAGE_TARGET_UNREACHABLE.format(error=e)

    _save_for_retry(
        status,
        error_message,
    )
    raise TargetNotAccessible()


def _try_send_webmention(
    status: OutgoingWebmentionStatus,
    source_urlpath: str,
    endpoint: str,
    target_url: str,
):
    success, status_code = _send_webmention(source_urlpath, endpoint, target_url)

    status.target_webmention_endpoint = endpoint
    status.response_code = status_code

    if success:
        log.info(f"Webmention submission successful for '{target_url}'")
        status.status_message = STATUS_MESSAGE_OK
        status.successful = True
        status.mark_processing_successful(save=True)
        return True

    else:
        log.warning(
            f"Webmention submission failed for '{target_url}' [endpoint={endpoint}]"
        )
        _save_for_retry(status, STATUS_MESSAGE_TARGET_ENDPOINT_ERROR)
        return False


def _send_webmention(
    source_urlpath: str,
    endpoint: str,
    target: str,
) -> Tuple[bool, int]:
    payload = {
        "target": target,
        "source": f"{options.url_scheme()}://{settings.DOMAIN_NAME}{source_urlpath}",
    }
    response = http_post(endpoint, data=payload)
    status_code = response.status_code
    if status_code >= 300:
        log.warning(
            f'Sending webmention to "{endpoint}" '
            f"FAILED with status_code={status_code}"
        )
        return False, status_code
    else:
        log.info(
            f'Sending webmention to "{endpoint}" '
            f"successful with status_code={status_code}"
        )
        return True, status_code


def _get_absolute_endpoint_from_response(response: Response) -> Optional[str]:
    endpoint = _get_endpoint_in_http_headers(
        response
    ) or _get_endpoint_in_html_response(response)
    abs_url = _relative_to_absolute_url(response, endpoint)
    log.debug(f"Absolute url: {endpoint} -> {abs_url}")
    return abs_url


def _get_endpoint_in_html_response(response: Response) -> Optional[str]:
    return _get_endpoint_in_html(response.text)


def _get_endpoint_in_http_headers(response: Response) -> Optional[str]:
    """Search for webmention endpoint in HTTP headers."""
    try:
        header_link = response.headers.get("Link")
        if "webmention" in header_link:
            endpoint = re.match(
                r'<(?P<url>.*)>[; ]*.rel=[\'"]?webmention[\'"]?', header_link
            ).group(1)
            log.debug(f"Webmention endpoint found in HTTP header: '{endpoint}'")
            return endpoint
    except Exception as e:
        log.debug(f"Unable to read HTTP headers: {e}")


def _get_endpoint_in_html(html_text: str) -> Optional[str]:
    """Search for a webmention endpoint in HTML."""
    a_soup = html_parser(html_text)

    # Check HTML <head> for <link> webmention endpoint
    try:
        links = a_soup.head.find_all("link", href=True, rel=True)
        for link in links:
            if "webmention" in link["rel"]:
                endpoint = link["href"]
                log.debug(f"Webmention endpoint found in document head: {endpoint}")
                return endpoint
    except Exception as e:
        log.debug(f"Error reading <head> of external link: {e}")

    # Check HTML <body> for <a> webmention endpoint
    try:
        links = a_soup.body.find_all("a", href=True, rel=True)
        for link in links:
            if "webmention" in link["rel"]:
                endpoint = link["href"]
                log.debug(f"Webmention endpoint found in document body: {endpoint}")
                return endpoint
    except Exception as e:
        log.debug(f"Error reading <body> of link: {e}")


def _relative_to_absolute_url(response: Response, url: str) -> Optional[str]:
    """
    If given url is relative, try to construct an absolute url using response domain.
    """
    if not url:
        return None

    try:
        URLValidator()(url)
        return url  # url is already well-formed.
    except ValidationError:
        scheme, domain, _, _, _ = urlsplit(response.url)
        if not scheme or not domain:
            return None
        return f"{scheme}://{domain}/" f'{url if not url.startswith("/") else url[1:]}'


def _save_for_retry(status: OutgoingWebmentionStatus, message: str) -> None:
    """In case of network errors, mark the status for reprocessing later."""
    log.warning(message)
    status.status_message = message
    status.mark_processing_failed(save=True)



from typing import Tuple
from urllib.parse import urlsplit

import requests
from bs4 import BeautifulSoup
from requests import Response

from mentions import options

__all__ = [
    "split_url",
    "html_parser",
    "http_get",
    "http_post",
]


HTTP_TIMEOUT_SECONDS = options.timeout()


def split_url(target_url: str) -> Tuple[str, str, str]:
    scheme, full_domain, path, _, _ = urlsplit(target_url)
    domain = full_domain.split(":")[0]  # Remove port number if present
    return scheme, domain, path


def html_parser(content) -> BeautifulSoup:
    return BeautifulSoup(content, features="html5lib")


def http_get(url: str) -> Response:
    return requests.get(url, timeout=HTTP_TIMEOUT_SECONDS)


def http_post(url: str, data: dict) -> Response:
    return requests.post(url, data=data, timeout=HTTP_TIMEOUT_SECONDS)



from typing import Dict, Iterable, List, Optional

from mentions.models import HCard, SimpleMention, Webmention
from mentions.models.mixins import QuotableMixin

__all__ = [
    "serialize_hcard",
    "serialize_mentions",
]


def serialize_mentions(mentions: Iterable[QuotableMixin]) -> List[Dict]:
    return [
        {
            "hcard": serialize_hcard(mention.hcard),
            "quote": mention.quote,
            "source_url": mention.source_url,
            "published": mention.published,
            "type": _typeof(mention),
        }
        for mention in mentions
    ]


def serialize_hcard(hcard: Optional[HCard]) -> Optional[Dict]:
    if hcard is None:
        return None

    return {
        "name": hcard.name,
        "avatar": hcard.avatar,
        "homepage": hcard.homepage,
    }


def _typeof(mention) -> str:
    if isinstance(mention, Webmention):
        return mention.post_type or "webmention"

    if isinstance(mention, SimpleMention):
        return "simple"

    raise ValueError(f"Unhandled mention type: {mention}")



from typing import List, Type

from django.db import models

from mentions.exceptions import ImplementationRequired
from mentions.models.mixins.quotable import QuotableMixin

__all__ = [
    "MentionableMixin",
]


class MentionableMixin(models.Model):
    class Meta:
        abstract = True

    allow_outgoing_webmentions = models.BooleanField(default=False)

    # `slug` field is no longer required as of version 2.3.0.
    # It may be removed in a future update, but for now we use it to provide
    # a default implementation of :func:`resolve_from_url_kwargs`
    slug = models.SlugField(unique=True)

    def mentions(self) -> List["QuotableMixin"]:
        from mentions.resolution import get_mentions_for_object

        return get_mentions_for_object(self)

    def mentions_json(self):
        from mentions.serialize import serialize_mentions

        return serialize_mentions(self.mentions())

    def get_absolute_url(self):
        raise ImplementationRequired(
            f"{self.__class__.__name__} does not implement get_absolute_url()"
        )

    def all_text(self) -> str:
        """
        Return all the text that should be searched when looking for
        outgoing Webmentions. Any URLs found in this text will be
        checked for webmention support.

        Typically this will just be the main text of your model, but
        you may also want to include content from any other text fields
        such as a summary or abstract.

        Example:
            def all_text(self) -> str:
                return f'{self.introduction} {self.main_content}'
        """
        raise ImplementationRequired(
            f"{self.__class__.__name__} does not implement all_text()"
        )

    @classmethod
    def resolve_from_url_kwargs(cls, **url_kwargs) -> Type["MentionableMixin"]:
        """Resolve a model instance from the given URL captured values.

        By default, an object is resolved via its unique slug.

        If your model does not have a unique slug, override this classmethod
        to retrieve an object using the URL parameters in url_kwargs.

        e.g.
        If your urlpatterns path looks like
            "<int:year>/<slug:slug>/",

        then you might override this method as:
            @classmethod
            def resolve_from_url_kwargs(cls, year: int, slug: str, **url_kwargs):
                return cls.objects.get(date__year=year, slug=slug)
        """
        return cls.objects.get(slug=url_kwargs.get("slug"))

    def should_process_webmentions(self) -> bool:
        """Return True if this instance should process webmentions when saved."""
        return self.allow_outgoing_webmentions

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

        if self.should_process_webmentions():
            from mentions.tasks import handle_outgoing_webmentions

            handle_outgoing_webmentions(self.get_absolute_url(), self.all_text())



from django.db import models

from mentions.models.base import MentionsBaseModel

__all__ = [
    "HCard",
]


class HCard(MentionsBaseModel):
    """An h-card represents information about a person or organisation.

    Spec: https://microformats.org/wiki/h-card
    """

    name = models.CharField(
        max_length=50,
        blank=True,
        null=True,
        help_text="Name of the person/organisation",
    )
    avatar = models.URLField(
        blank=True,
        null=True,
        help_text="Link to their profile image",
    )
    homepage = models.URLField(
        blank=True,
        null=True,
        help_text="Link to their homepage",
    )
    json = models.TextField(
        blank=True,
        help_text="Raw json representation of this hcard",
    )

    def __str__(self):
        return f'HCard(name="{self.name}", avatar="{self.avatar}", homepage="{self.homepage})"'



from django.db import models

from mentions.models.base import MentionsBaseModel
from mentions.models.mixins import RetryableMixin

__all__ = [
    "OutgoingWebmentionStatus",
]


class OutgoingWebmentionStatus(RetryableMixin, MentionsBaseModel):
    """Status tracker for webmentions that you (attempt to) send from your server.

    Used primarily for logging of outgoing mentions.
    """

    source_url = models.URLField(
        help_text="The URL on your server where this mention originates",
    )
    target_url = models.URLField(
        help_text="The URL that you mentioned.",
    )
    target_webmention_endpoint = models.URLField(
        null=True,
        blank=True,
        help_text="The endpoint URL to which we sent the webmention",
    )
    status_message = models.CharField(
        max_length=1024,
        help_text="Success, or an explanation of what went wrong.",
    )
    response_code = models.PositiveIntegerField(default=0)

    successful = models.BooleanField(default=False)

    def __str__(self):
        return (
            f"[{'OK' if self.successful else 'Failed'}:{self.response_code}] "
            f"{self.source_url} -> {self.target_url}"
        )

    class Meta:
        ordering = ["-created_at"]
        verbose_name_plural = "Outgoing Webmentions"



import logging
from importlib import import_module
from typing import Iterable, List, Optional, Type

from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from django.http import HttpRequest
from django.urls import Resolver404, ResolverMatch

from mentions.exceptions import BadConfig, TargetDoesNotExist
from mentions.models import HCard, OutgoingWebmentionStatus, SimpleMention, Webmention
from mentions.models.base import MentionsBaseModel
from mentions.models.mixins import MentionableMixin, QuotableMixin
from mentions.util import split_url

log = logging.getLogger(__name__)


__all__ = [
    "get_mentions_for_absolute_url",
    "get_mentions_for_object",
    "get_mentions_for_url_path",
    "get_mentions_for_view",
    "get_model_for_url_path",
]


def _find_urlpattern(target_path: str) -> ResolverMatch:
    """
    Find a match in urlpatterns or raise TargetDoesNotExist
    """
    target_path = target_path.lstrip("/")  # Remove any leading slashes
    urlconf = import_module(settings.ROOT_URLCONF)
    urlpatterns = urlconf.urlpatterns

    for x in urlpatterns:
        # x may be an instance of either:
        # - django.urls.resolvers.URLResolver
        # - django.urls.resolvers.URLPattern
        try:
            match = x.resolve(target_path)
            if match:
                break
        except Resolver404:
            # May be raised by URLResolver
            pass
    else:
        # No match found
        raise TargetDoesNotExist(
            f"Cannot find a matching urlpattern entry for path={target_path}"
        )
    log.debug(f"Found matching urlpattern: {match}")
    return match


def get_model_for_url_path(
    target_path: str,
    match: ResolverMatch = None,
) -> Type[MentionableMixin]:
    """
    Find a match in urlpatterns and return the corresponding model instance.

    Any mentionable View that corresponds to a model instance should include
    kwargs in its `urlpatterns` definition to enable resolution.
    - `model_name`: dotted python path for the target model.
    - Any other kwargs are passed to the model's `resolve_from_url_kwargs`
      classmethod to resolve the model instance. The default implementation
      uses the `slug` kwarg but this may be overridden.

        e.g.
        path(
            r"articles/<slug:slug>/",
            ArticleView.as_view(),
            kwargs={
                "model_name": "blog.Article",
            },
            name="blog-article"),
    """
    if match is None:
        match = _find_urlpattern(target_path)

    urlpath_kwargs: dict = dict(**match.kwargs)

    # Dotted path to model class declaration
    try:
        model_name = urlpath_kwargs.pop("model_name")
    except KeyError:
        raise BadConfig(
            f"urlpattern must include a kwargs entry called 'model_name': {match}"
        )

    try:
        model = apps.get_model(model_name)
    except LookupError:
        raise BadConfig(f"Cannot find model `{model_name}` - check your urlpattern!")

    try:
        return model.resolve_from_url_kwargs(**urlpath_kwargs)
    except ObjectDoesNotExist:
        raise TargetDoesNotExist(
            f"Cannot find instance of model='{model}' with kwargs='{urlpath_kwargs}'"
        )


def get_mentions_for_url_path(
    target_path: str,
    full_target_url: str,
) -> List[QuotableMixin]:
    """If target_path resolves to a page associated with a MentionableMixin model"""
    match = _find_urlpattern(target_path)

    try:
        obj = get_model_for_url_path(target_path, match)
        return obj.mentions()
    except (BadConfig, TargetDoesNotExist):
        pass

    # Add or remove a trailing slash to full_target_url so we can look for both.
    full_target_url_invert_slash = (
        full_target_url[:-1] if full_target_url.endswith("/") else f"{full_target_url}/"
    )

    q_filter = (
        Q(target_url=full_target_url)
        | Q(target_url=full_target_url_invert_slash)
        | Q(target_url=target_path)
    )

    webmentions = Webmention.objects.filter(
        q_filter,
        approved=True,
        validated=True,
    )
    simple_mentions = SimpleMention.objects.filter(q_filter)

    return list(webmentions) + list(simple_mentions)


def get_mentions_for_absolute_url(url: str) -> Iterable[QuotableMixin]:
    scheme, domain, path = split_url(url)
    full_target_url = f"{scheme}://{domain}{path}"

    return get_mentions_for_url_path(path, full_target_url=full_target_url)


def get_mentions_for_view(request: HttpRequest) -> Iterable[QuotableMixin]:
    """Call from your View implementation so you can include mentions in your template."""

    return get_mentions_for_absolute_url(request.build_absolute_uri())


def get_mentions_for_object(obj: MentionableMixin) -> List[QuotableMixin]:
    ctype = ContentType.objects.get_for_model(obj.__class__)
    webmentions = Webmention.objects.filter(
        content_type=ctype,
        object_id=obj.id,
        approved=True,
        validated=True,
    )
    simple_mentions = SimpleMention.objects.filter(
        content_type=ctype,
        object_id=obj.id,
    )
    return list(webmentions) + list(simple_mentions)


def get_or_create_outgoing_webmention(
    source_urlpath: str,
    target_url: str,
    reset_retries: bool = False,
) -> OutgoingWebmentionStatus:
    """Safely get or create a OutgoingWebmentionStatus instance for given URLs.

    Since version 3.0.0 the URLs are effectively treated as 'unique together',
    but there may exist 'duplicate' instances from a previous installation."""

    kwargs = {
        "source_url": source_urlpath,
        "target_url": target_url,
    }
    try:
        status = OutgoingWebmentionStatus.objects.get_or_create(**kwargs)[0]
    except OutgoingWebmentionStatus.MultipleObjectsReturned:
        status = OutgoingWebmentionStatus.objects.filter(**kwargs).first()

    if reset_retries:
        status.reset_retries()

    return status


def update_or_create_hcard(
    homepage: Optional[str],
    name: Optional[str],
    avatar: Optional[str],
    data: str,
) -> HCard:
    """Any individual field may be used to create/retrieve an HCard.

    Ideally, homepage and name are used together.

    Otherwise, the order of precedence is [homepage, name, avatar].
    """

    if homepage and name:
        return _update_first_or_create(
            HCard,
            homepage=homepage,
            name=name,
            defaults={
                "avatar": avatar,
                "json": data,
            },
        )

    if homepage:
        return _update_first_or_create(
            HCard,
            homepage=homepage,
            name=None,
            defaults={
                "avatar": avatar,
                "json": data,
            },
        )

    if name:
        return _update_first_or_create(
            HCard,
            homepage=None,
            name=name,
            defaults={
                "avatar": avatar,
                "json": data,
            },
        )

    if avatar:
        return _update_first_or_create(
            HCard,
            homepage=None,
            name=None,
            avatar=avatar,
            defaults={
                "json": data,
            },
        )


def _update_first_or_create(
    model_cls: Type[MentionsBaseModel],
    defaults: dict,
    **query,
):
    try:
        return model_cls.objects.update_or_create(
            **query,
            defaults=defaults,
        )[0]

    except model_cls.MultipleObjectsReturned:
        instance = model_cls.objects.filter(**query).first()
        for key, value in defaults.items():
            setattr(instance, key, value)

        instance.save()
        return instance



from django.contrib import admin

from mentions.models import (
    HCard,
    OutgoingWebmentionStatus,
    PendingIncomingWebmention,
    PendingOutgoingContent,
    SimpleMention,
    Webmention,
)


@admin.action(permissions=["change"])
def approve_webmention(modeladmin, request, queryset):
    queryset.update(approved=True)


@admin.action(permissions=["change"])
def disapprove_webmention(modeladmin, request, queryset):
    queryset.update(approved=False)


class BaseAdmin(admin.ModelAdmin):
    save_on_top = True


@admin.register(SimpleMention)
class QuotableAdmin(BaseAdmin):
    list_display = [
        "source_url",
        "target_url",
        "hcard",
    ]
    search_fields = [
        "source_url",
        "target_url",
        "hcard",
    ]
    readonly_fields = [
        "target_object",
        "published",
    ]
    date_hierarchy = "published"


@admin.register(Webmention)
class WebmentionAdmin(QuotableAdmin):
    readonly_fields = QuotableAdmin.readonly_fields + [
        "content_type",
        "object_id",
    ]
    actions = [
        approve_webmention,
        disapprove_webmention,
    ]
    list_display = [
        "source_url",
        "target_url",
        "published",
        "validated",
        "approved",
        "target_object",
    ]
    fieldsets = (
        (
            "Remote source",
            {
                "fields": (
                    "source_url",
                    "sent_by",
                    "hcard",
                    "quote",
                    "post_type",
                ),
            },
        ),
        (
            "Local target",
            {
                "fields": (
                    "target_url",
                    "content_type",
                    "object_id",
                    "target_object",
                ),
            },
        ),
        (
            "Metadata",
            {
                "fields": (
                    "published",
                    "approved",
                    "validated",
                    "notes",
                ),
            },
        ),
    )


@admin.register(OutgoingWebmentionStatus)
class OutgoingWebmentionStatusAdmin(BaseAdmin):
    readonly_fields = [
        "created_at",
        "source_url",
        "target_url",
        "target_webmention_endpoint",
        "status_message",
        "response_code",
        "successful",
    ]
    list_display = [
        "source_url",
        "target_url",
        "successful",
        "created_at",
    ]
    date_hierarchy = "created_at"


@admin.register(HCard)
class HCardAdmin(BaseAdmin):
    list_display = ["name", "homepage"]
    search_fields = ["name", "homepage"]


@admin.register(PendingIncomingWebmention)
class PendingIncomingAdmin(BaseAdmin):
    pass


@admin.register(PendingOutgoingContent)
class PendingOutgoingAdmin(BaseAdmin):
    pass



from typing import Set

from mentions.util import html_parser

__all__ = [
    "get_target_links_in_text",
]


def get_target_links_in_text(text: str) -> Set[str]:
    """Get any links from the text that should be treated as webmention targets."""
    links = _find_links_in_text(text)

    # Filter self-linking #anchors.
    links = {link for link in links if not link.startswith("#")}

    return links


def _find_links_in_text(text: str) -> Set[str]:
    """Get the raw target href of any links in the text."""
    soup = html_parser(text)
    return {a["href"] for a in soup.find_all("a", href=True)}



from django.urls import path, re_path

from mentions.views import (
    GetMentionsView,
    WebmentionDashboardView,
    WebmentionView,
    view_names,
)

"""
/webmention/
Configure in root project urls.py
e.g.
    from django.conf import settings
    urlpatterns = [
        ...,
        path(r'webmention/', include('mentions.urls')),
    ]
"""
urlpatterns = [
    path("", WebmentionView.as_view(), name=view_names.webmention_api_incoming),
    re_path(
        r"^get/?$",
        GetMentionsView.as_view(),
        name=view_names.webmention_api_get_for_object,
    ),
    path(
        "dashboard/",
        WebmentionDashboardView.as_view(),
        name=view_names.webmention_dashboard,
    ),
]



import logging

from django.http import HttpResponseBadRequest, JsonResponse
from django.views import View

from mentions.exceptions import TargetDoesNotExist
from mentions.resolution import get_mentions_for_url_path
from mentions.serialize import serialize_mentions
from mentions.util import split_url

__all__ = [
    "GetMentionsView",
]

log = logging.getLogger(__name__)


# /webmention/get
class GetMentionsView(View):
    """Return any mentions (Webmention, SimpleMention) associated with a given url."""

    def get(self, request):
        for_url = request.GET.get("url")

        if not for_url:
            return HttpResponseBadRequest("Missing args")

        scheme, domain, path = split_url(request.build_absolute_uri())
        full_target_url = f"{scheme}://{domain}{for_url}"

        try:
            wm = get_mentions_for_url_path(for_url, full_target_url=full_target_url)

            return JsonResponse(
                {
                    "target_url": full_target_url,
                    "mentions": serialize_mentions(wm),
                }
            )

        except TargetDoesNotExist as e:
            log.warning(e)
            return JsonResponse(
                {
                    "target_url": full_target_url,
                    "message": "Target not found",
                    "mentions": [],
                },
                status=404,
            )



from django.db import models

__all__ = [
    "MentionsBaseModel",
]


class MentionsBaseModel(models.Model):
    created_at = models.DateTimeField(auto_now_add=True, null=True)

    class Meta:
        abstract = True



from django.db import models

from mentions import options
from mentions.models.base import MentionsBaseModel
from mentions.models.mixins import QuotableMixin

__all__ = [
    "Webmention",
]


def _approve_default():
    return options.auto_approve()


class Webmention(QuotableMixin, MentionsBaseModel):
    """An incoming webmention that is received by your server."""

    sent_by = models.URLField(
        blank=True,
        help_text="Source address of the HTTP request that sent this webmention",
    )

    approved = models.BooleanField(
        default=_approve_default, help_text="Allow this webmention to appear publicly"
    )
    validated = models.BooleanField(
        default=False,
        help_text="True if both source and target have been validated, "
        "confirmed to exist, and source really does link to target",
    )

    notes = models.CharField(
        max_length=1024,
        blank=True,
        help_text="A description of any errors encountered when building this Webmention.",
    )

    def approve(self):
        self.approved = True

    class Meta:
        ordering = ["-created_at"]

    def __str__(self):
        return (
            f"{self.source_url} -> {self.target_url} "
            f"[validated={self.validated}, approved={self.approved},"
            f"content_type={self.content_type}, id={self.object_id}]"
        )



from mentions.exceptions import (
    SourceDoesNotLink,
    SourceNotAccessible,
    TargetDoesNotExist,
    TargetWrongDomain,
)
from mentions.models import PendingIncomingWebmention, Webmention
from mentions.tasks.celeryproxy import get_logger, shared_task
from mentions.tasks.incoming.local import get_target_object
from mentions.tasks.incoming.remote import get_metadata_from_source, get_source_html

__all__ = [
    "process_incoming_webmention",
]

log = get_logger(__name__)


@shared_task
def process_incoming_webmention(source_url: str, target_url: str, sent_by: str) -> None:
    log.info(f"Processing webmention '{source_url}' -> '{target_url}'")

    wm = Webmention(source_url=source_url, target_url=target_url, sent_by=sent_by)
    notes = _Notes()

    try:
        wm.target_object = get_target_object(target_url)

    except (TargetWrongDomain, TargetDoesNotExist):
        notes.warning(
            f"Unable to find matching page on our server for url '{target_url}'"
        )

    try:
        response_html = get_source_html(source_url)

    except SourceNotAccessible:
        return _save_for_retry(source_url, target_url, sent_by)

    try:
        wm = _get_metadata(wm, response_html, target_url)
        _save_mention(wm, notes, validated=True)

    except SourceDoesNotLink:
        _save_mention(
            wm,
            notes=notes.warning("Source does not contain a link to our content"),
        )


class _Notes:
    """Keep track of any issues that might need to be checked manually."""

    notes = []

    def warning(self, note) -> "_Notes":
        log.warning(note)
        self.notes.append(note)
        return self

    def __str__(self):
        return "\n".join(self.notes)[:1023]


def _save_mention(wm: Webmention, notes: _Notes, validated: bool = False):
    wm.notes = str(notes)
    wm.validated = validated
    wm.save()

    try:
        pending = PendingIncomingWebmention.objects.get(
            target_url=wm.target_url,
            source_url=wm.source_url,
        )
        pending.mark_processing_successful(save=True)
    except PendingIncomingWebmention.DoesNotExist:
        pass


def _save_for_retry(source_url: str, target_url: str, sent_by: str):
    """In case of network errors, create or update PendingIncomingWebmention instance to retry later."""
    log.info(f"_save_for_retry {source_url} -> {target_url}")

    pending, _ = PendingIncomingWebmention.objects.get_or_create(
        target_url=target_url,
        source_url=source_url,
        defaults={
            "sent_by": sent_by,
        },
    )

    pending.mark_processing_failed(save=True)


def _get_metadata(mention: Webmention, html: str, target_url: str) -> Webmention:
    metadata = get_metadata_from_source(html=html, target_url=target_url)
    mention.post_type = metadata.post_type
    mention.hcard = metadata.hcard
    return mention
